### DYNAMIC CITY NAMES
### Version 1.0
### By Xyth

### No changes need to be made to this file when merging

### Imports
from CvPythonExtensions import *
import copy



### Globals
gc = CyGlobalContext()
CityDict = {}
PlayerDict = {}



### Classes

class CityInfo:
	def __init__(self):
		self.CityName = ""
		self.Civilizations = {}
		self.Leaders = {}
		self.Eras = {}



class CityNameInfo:
	def __init__(self):
		self.CityName = ""
		self.Capital = False
		self.Priority = 0



### Functions

def initDatabase(CityFile, CityListFile, bDebug):
	'Initialises the city and player dictionaries'
	global CityDict
	global PlayerDict

	readCityInfos(CityFile)
	readCityListInfos(CityListFile)

	if bDebug == True:
		print ""
		print "City Dictionary (All Cities)"
		printCityInfo(CityDict)

	buildPlayerDict()

	if bDebug == True:
		for iPlayer in range(gc.getMAX_PLAYERS()):
			pPlayer = gc.getPlayer(iPlayer)
			if pPlayer.isAlive():
				print ""
				print "City Dictionary for Player " + str(iPlayer) + " (" + pPlayer.getCivilizationDescription(0) +")"
				for key, info in PlayerDict[iPlayer].iteritems():
					printCityNameInfo(" - ", key, info)



def readCityInfos(file):
	'Loads all CityInfo entries from a file into the city dictionary'
	global CityDict
	for line in file.readlines():

		# Setup new city entry
		if "<CityName>" in line:
			city = CityInfo()
			key = createCityKey(getXMLString(line))
			city.CityName = getXMLString(line).decode('utf-8')

		# Parse Civilization info
		elif "<CivilizationType>" in line:
			civ = CityNameInfo()
			type = gc.getInfoTypeForString(getXMLString(line))
			if type == -1:
				print "DCN: Error! " + str(type) + " is not a valid Civilization"
		elif "<CivilizationCityName>" in line:
			civ.CityName = getXMLString(line)
		elif "<bCivilizationCapital>" in line:
			if int(getXMLString(line)) == 1:
				civ.Capital = True
		elif "<iCivilizationPriority>" in line:
			civ.Priority = int(getXMLString(line))
		elif "</Civilization>" in line:
			city.Civilizations[type] = civ

		# Parse Leader info
		elif "<LeaderType>" in line:
			leader = CityNameInfo()
			type = gc.getInfoTypeForString(getXMLString(line))
			if type == -1:
				print "DCN: Error! " + str(type) + " is not a valid Leader"
		elif "<LeaderCityName>" in line:
			leader.CityName = getXMLString(line).decode('utf-8')
		elif "<bLeaderCapital>" in line:
			if int(getXMLString(line)) == 1:
				leader.Capital = True
		elif "<iLeaderPriority>" in line:
			leader.Priority = int(getXMLString(line))
		elif "</Leader>" in line:
			city.Leaders[type] = leader

		# Parse Era info
		elif "<EraType>" in line:
			era = CityNameInfo()
			type = gc.getInfoTypeForString(getXMLString(line))
			if type == -1:
				print "DCN: Error! " + str(type) + " is not a valid Era"
		elif "<EraCityName>" in line:
			era.CityName = getXMLString(line).decode('utf-8')
		elif "<bEraCapital>" in line:
			if int(getXMLString(line)) == 1:
				era.Capital = True
		elif "<iEraPriority>" in line:
			era.Priority = int(getXMLString(line))
		elif "</Era>" in line:
			city.Eras[type] = era

		# Add city entry to dictionary
		elif "</CityInfo>" in line:
			CityDict[key] = city



def readCityListInfos(file):
	'Loads all CityListInfo entries from a file into the city dictionary'
	global CityDict
	for line in file.readlines():

		# Set up city entry and list of names
		if "<CityNames>" in line:
			namelist = []
			city = CityInfo()
		elif "<CityName>" in line:
			namelist.append(getXMLString(line).decode('utf-8'))

		# Parse Civilization info
		elif "<CivilizationType>" in line:
			civ = CityNameInfo()
			civtype = gc.getInfoTypeForString(getXMLString(line))
			if type == -1:
				print "DCN: Error! " + str(type) + " is not a valid Civilization"
		elif "<iCivilizationPriority>" in line:
			civ.Priority = int(getXMLString(line))
		elif "</Civilization>" in line:
			city.Civilizations[civtype] = civ

		# Parse Leader info
		elif "<LeaderType>" in line:
			leader = CityNameInfo()
			leadertype = gc.getInfoTypeForString(getXMLString(line))
			if type == -1:
				print "DCN: Error! " + str(type) + " is not a valid Leader"
		elif "<iLeaderPriority>" in line:
			leader.Priority = int(getXMLString(line))
		elif "</Leader>" in line:
			city.Leaders[leadertype] = leader

		# Parse Era info
		elif "<EraType>" in line:
			era = CityNameInfo()
			eratype = gc.getInfoTypeForString(getXMLString(line))
			if type == -1:
				print "DCN: Error! " + str(type) + " is not a valid Era"
		elif "<iEraPriority>" in line:
			era.Priority = int(getXMLString(line))
		elif "</Era>" in line:
			city.Eras[eratype] = era

		# Add each city entry to dictionary
		elif "</CityListInfo>" in line:
			for iName in range(len(namelist)):
				key = createCityKey(namelist[iName])
				CityDict[key] = copy.deepcopy(city)
				CityDict[key].CityName = namelist[iName]
				for iCiv, civinfo in CityDict[key].Civilizations.iteritems():
					if civinfo.CityName == "":
						CityDict[key].Civilizations[iCiv].CityName = namelist[iName]
				for iLeader, leaderinfo in CityDict[key].Leaders.iteritems():
					if leaderinfo.CityName == "":
						CityDict[key].Leaders[iLeader].CityName = namelist[iName]
				for iEra, erainfo in CityDict[key].Eras.iteritems():
					if erainfo.CityName == "":
						CityDict[key].Eras[iEra].CityName = namelist[iName]
			del namelist[:]



def buildPlayerDict():
	'Creates a dictionary of valid cities for each player'
	global CityDict
	global PlayerDict
	FoundedList = []

	for iPlayer in range(gc.getMAX_PLAYERS()):
		pPlayer = gc.getPlayer(iPlayer)
		if pPlayer.isAlive():
			PlayerDict[iPlayer] = {}
			for iCity in range(pPlayer.getNumCities()):
				foundedkey = getCityKey(pPlayer.getCity(iCity).getName(), iPlayer)
				FoundedList.append(foundedkey)

	for key in CityDict.keys():
		if not key in FoundedList:
			enableCityName(key)
		else:
			print "DCN: " + key + " already exists in game"



def createCityKey(cityname):
	'Creates a dictionary key from cityname'
	ascii = "".join([x for x in cityname if ord(x) < 128])
	key = "CITY_" + ascii.upper()
	return key



def getCityKey(cityname, iPlayer):
	'Returns the default name for a city'
	global CityDict
	global PlayerDict

	# Check if the name is already the default
	citykey = createCityKey(cityname)
	if citykey in CityDict.keys():
		return citykey

	# Look for default name by civilization or leader
	for key, info in PlayerDict[iPlayer].iteritems():
		if info.CityName == cityname:
			return CityDict[key].CityName

	# Look for default name by era
	for key, info in CityDict.iteritems():
		for iEra in range(gc.getNumEraInfos()):
			if iEra in info.Eras.keys():
				if info.Eras[iEra].CityName == cityname:
					return CityDict[key].CityName

	return ""



def enableCityName(key):
	'Enables founding of a city for relevent players'
	global CityDict
	global PlayerDict

	# Loop through valid players
	for iPlayer in range(gc.getMAX_PLAYERS()):
		pPlayer = gc.getPlayer(iPlayer)
		if pPlayer.isAlive():
			iCiv = pPlayer.getCivilizationType()
			iLeader = pPlayer.getLeaderType()
			if key in CityDict.keys():

				# Try to use Leader info first
				if iLeader in CityDict[key].Leaders:
					info = CityDict[key].Leaders[iLeader]
					if iCiv in CityDict[key].Civilizations:
						# Add priorities
						info.Priority += CityDict[key].Civilizations[iCiv].Priority

				# Then try to use Civilization info
				elif iCiv in CityDict[key].Civilizations:
					info = CityDict[key].Civilizations[iCiv]

				# Otherwise city is not valid for this player
				else:
					continue

				# Add city to player dictionary
				if info.Priority > 0:
					PlayerDict[iPlayer][key] = info
					print "DCN: " + key + " added to citylist of Player " + str(iPlayer) + " (" + gc.getPlayer(iPlayer).getCivilizationDescription(0) +")"



def disableCityName(key):
	'Disables founding of a city for all players'
	global PlayerDict

	for iPlayer in PlayerDict.keys():
		if key in PlayerDict[iPlayer]:
			del PlayerDict[iPlayer][key]
			print "DCN: " + key + " removed from citylist of Player " + str(iPlayer) + " (" + gc.getPlayer(iPlayer).getCivilizationDescription(0) +")"



def selectCityName(iPlayer, bCapital):
	'Selects a city the player can found based on era-adjusted priority'
	global CityDict
	global PlayerDict
	SelectionList = []
	highestpriority = 0
	iEra = gc.getPlayer(iPlayer).getCurrentEra()

	for key in PlayerDict[iPlayer].keys():
		# If specified, only consider capitals
		if bCapital:
			if isCapital(key, iPlayer) == False:
				continue

		# Adjust priority for current era
		priority = PlayerDict[iPlayer][key].Priority
		if iEra in CityDict[key].Eras:
			priority += CityDict[key].Eras[iEra].Priority

		# Determine highest priority
		if priority > highestpriority:
			highestpriority = priority

	for key in PlayerDict[iPlayer].keys():
		# If specified, only consider capitals
		if bCapital:
			if isCapital(key, iPlayer) == False:
				continue

		#  Add highest priority cities to selection list
		priority = PlayerDict[iPlayer][key].Priority
		if iEra in CityDict[key].Eras:
			priority += CityDict[key].Eras[iEra].Priority
		if priority == highestpriority:
			SelectionList.append(key)

	if len(SelectionList) == 0:
		# Abort if no cities are eligible (fall back on default system)
		print ""
		print "DCN: No eligible cities found for Player " + str(iPlayer) + ". Standard city naming system used."
		return ""

	else:
		# Otherwise choose a random city from selection list
		rand = CyGame().getSorenRandNum(len(SelectionList), "City Name Selection")
		selection = SelectionList[rand]
		cityname = PlayerDict[iPlayer][selection].CityName

		# Check if city should have a different name in current era
		if iEra in CityDict[selection].Eras:
			if CityDict[selection].Eras[iEra].CityName != "":
				if CityDict[selection].Eras[iEra].CityName != cityname:
					cityname = CityDict[selection].Eras[iEra].CityName

		name = getOutputEncoding(cityname)

		print ""
		print "DYNAMIC CITY NAMES"
		print "Priority: " + str(highestpriority)
		print "Candidates: " + str(SelectionList)
		print selection + " founded by Player " + str(iPlayer) + " as " + name

		disableCityName(selection)
		return unicode(cityname, 'utf-8')



def renameCityName(iPlayer, oldname, newname):
	'Adjusts player dictionary when a city is renamed'
	oldkey = getCityKey(oldname, iPlayer)
	newkey = getCityKey(newname, iPlayer)
	enableCityName(oldkey)
	disableCityName(newkey)



def isCapital(key, iPlayer):
	'Returns true if city is a suitable capital'
	global CityDict
	global PlayerDict
	iEra = gc.getPlayer(iPlayer).getCurrentEra()

	if key in PlayerDict[iPlayer]:
		if PlayerDict[iPlayer][key].Capital == True:
			return True
		elif iEra in CityDict[key].Eras:
			if CityDict[key].Eras[iEra].Capital == True:
				return True

	return False



### Utility Functions

def getXMLString(string):
	'Returns the content found between XML tags as a string'
	string = str(string)
	string = string.strip()
	string = string[2:-1]
	BeginPos = -1
	EndPos = -1
	for i in xrange(len(string)):
		if string[i] == ">":
			BeginPos = i
		elif string[i] == "<":
			EndPos = i
			break
	else:
		return "-1"
	NewString = string[BeginPos + 1 : EndPos]
	return str(NewString)



def getOutputEncoding(string):
	'Encodes string correctly for console output'
	if type(string) == str:
		return string
	else:
		return string.encode('utf-8')



def printCityInfo(dict):
	'Outputs the contents of a city dictionary for debugging'
	for key in dict.keys():
		print ""
		print key
		for civkey, civinfo in dict[key].Civilizations.iteritems():
			printCityNameInfo("Civilization:", civkey, civinfo)
		for leaderkey, leaderinfo in dict[key].Leaders.iteritems():
			printCityNameInfo("Leader:      ", leaderkey, leaderinfo)
		for erakey, erainfo in dict[key].Eras.iteritems():
			printCityNameInfo("Era:         ", erakey, erainfo)
	print ""



def printCityNameInfo(title, key, info):
	'Outputs city information for debugging'
	if info.CityName == "":
		info.CityName = "None"
	name = getOutputEncoding(info.CityName)
	print title + " " + str(key) + " (Name = " + name + ", Capital = " + str(info.Capital) + ", Priority = " + str(info.Priority) + ")"



